<?php

namespace wpbuddy\rich_snippets;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
} // Exit if accessed directly


/**
 * Class Admin_Snippets_Forms_Controller.
 *
 * Manages creation and processing of Rich Snippets that can be overwritten.
 *
 * @package wpbuddy\rich_snippets
 *
 * @since   2.2.0
 */
final class Admin_Snippets_Overwrite_Controller {

	/**
	 * The instance.
	 *
	 * @var Admin_Snippets_Overwrite_Controller
	 *
	 * @since 2.2.0
	 */
	protected static $instance = null;


	/**
	 * If this instance has been initialized already.
	 *
	 * @since 2.2.0
	 *
	 * @var bool
	 */
	protected $initialized = false;


	/**
	 * Get the singleton instance.
	 *
	 * Creates a new instance of the class if it does not exists.
	 *
	 * @return   Admin_Snippets_Overwrite_Controller
	 *
	 * @since 2.2.0
	 */
	public static function instance() {

		if ( null === self::$instance ) {
			self::$instance = new self;
		}

		return self::$instance;
	}


	/**
	 * Magic function for cloning.
	 *
	 * Disallow cloning as this is a singleton class.
	 *
	 * @since 2.2.0
	 */
	protected function __clone() {
	}


	/**
	 * Magic method for setting up the class.
	 *
	 * Disallow external instances.
	 *
	 * @since 2.2.0
	 */
	protected function __construct() {
	}


	/**
	 * Init.
	 *
	 * @since 2.2.0
	 */
	public function init() {

		if ( $this->initialized ) {
			return;
		}

		add_action( 'admin_footer', [ self::instance(), 'print_modal_window' ] );

		$this->add_metaboxes();

		$this->initialized = true;
	}


	/**
	 * @since 2.2.0
	 */
	public function print_modal_window() {

		View::admin_posts_modalwindow();
	}


	/**
	 * Add metaboxes.
	 *
	 * @since 2.2.0
	 */
	public function add_metaboxes() {

		if ( ! Helper_Model::instance()->magic() ) {
			return;
		}

		add_meta_box(
			'rswp-overwrite-mb',
			__( 'Global Snippets Values', 'rich-snippets-schema' ),
			array( '\wpbuddy\rich_snippets\View', 'admin_posts_metabox_overwrites' ),
			null,
			'side',
			'high'
		);
	}


	/**
	 * Back Compat for Overwrite data.
	 * In version 2.14.0 we changed the way how overwrite data is stored. We correct that here.
	 *
	 * @param array $data
	 *
	 * @return array
	 *
	 * @since      2.14.0
	 */
	private function back_compat_overwrite( $data, $post_id ) {
		if ( ! is_array( $data ) ) {
			return [];
		}

		if ( count( $data ) <= 0 ) {
			return $data;
		}

		$new_data = [];

		foreach ( $data as $snip_id => $d ) {
			if ( ! isset( $d['id'] ) ) {
				$new_data[ $snip_id ] = $d;
				continue;
			}

			if ( ! isset( $d['properties'] ) ) {
				continue;
			}

			$post_id = Snippets_Model::get_post_id_by_snippet_id( $snip_id );

			if ( ! $post_id ) {
				continue;
			}

			$snippet = Snippets_Model::get_first_snippet( $post_id );

			if ( ! $snippet ) {
				continue;
			}

			if ( $snippet->id !== $snip_id ) {
				continue;
			}

			foreach ( $d['properties'] as $property_id => $property_value ) {
				if ( ! is_array( $property_value ) ) {
					$property_value = [ $property_value ];
				}

				foreach ( $property_value as $pv ) {
					$new_data[ $snippet->id ][ uniqid() ] = [
						"prop_id"           => $property_id,
						"snippet_id"        => $snippet->id,
						"parent_snippet_id" => $snippet->id,
						"value"             => $pv
					];
				}

				unset( $property_id, $property_value );
			}

			foreach ( $snippet->get_sub_snippet_ids_deep() as $property_name_and_id => $sub_snippet_id ) {
				if ( ! isset( $data[ $sub_snippet_id ] ) ) {
					continue;
				}

				$sub_snip = $data[ $sub_snippet_id ];

				if ( ! isset( $sub_snip['properties'] ) ) {
					continue;
				}

				foreach ( $sub_snip['properties'] as $property_id => $property_value ) {
					if ( ! is_array( $property_value ) ) {
						$property_value = [ $property_value ];
					}

					foreach ( $property_value as $pv ) {
						$new_data[ $snippet->id ][ uniqid() ] = [
							"prop_id"           => $property_id,
							"snippet_id"        => $sub_snippet_id,
							"parent_snippet_id" => $snippet->id,
							"value"             => $pv
						];
					}

					unset( $property_id, $property_value );
				}
			}

		}

		if ( md5( serialize( $new_data ) ) !== md5( serialize( $data ) ) ) {
			update_post_meta( $post_id, '_wpb_rs_overwrite_data', $new_data );
		}

		return $new_data;
	}

	/**
	 * Fetches overwrite data for a main snippet.
	 *
	 * @param int    $post_id
	 * @param string $main_snippet
	 * @param string $snippet_id
	 * @param string $parent_snippet_id
	 *
	 * @return array
	 *
	 * @since 2.14.0
	 */
	private function get_properties_to_overwrite( $post_id, $main_snippet, $snippet_id, $parent_snippet_id ) {
		$overwrite_data = get_post_meta( $post_id, '_wpb_rs_overwrite_data', true );
		$overwrite_data = $this->back_compat_overwrite( $overwrite_data, $post_id );

		if ( ! is_array( $overwrite_data ) ) {
			return [];
		}

		if ( count( $overwrite_data ) <= 0 ) {
			return [];
		}

		if ( ! isset( $overwrite_data[ $main_snippet ] ) ) {
			return [];
		}

		$d = [];

		foreach ( $overwrite_data[ $main_snippet ] as $prop ) {
			if ( $prop['snippet_id'] === $snippet_id && $prop['parent_snippet_id'] == $parent_snippet_id ) {
				$d[] = $prop;
			}
		}

		return $d;
	}


	/**
	 * Fetches extra properties for list elements.
	 *
	 * @param int          $post_id
	 * @param string       $main_snippet
	 * @param Rich_Snippet $snippet
	 *
	 * @return array
	 *
	 * @since 2.14.0
	 */
	private function get_properties_to_list( $post_id, $main_snippet, $snippet ) {
		$overwrite_data = get_post_meta( $post_id, '_wpb_rs_overwrite_data', true );
		$overwrite_data = $this->back_compat_overwrite( $overwrite_data, $post_id );

		if ( ! is_array( $overwrite_data ) ) {
			return [];
		}

		if ( count( $overwrite_data ) <= 0 ) {
			return [];
		}

		if ( ! isset( $overwrite_data[ $main_snippet ] ) ) {
			return [];
		}

		$snippet_ids = $snippet->get_sub_snippet_ids();

		$d = [];

		$ends_with_fct = function ( $haystack, $needle ) {
			return substr_compare( $haystack, $needle, - strlen( $needle ) ) === 0;
		};

		foreach ( $snippet_ids as $prop_name => $snippet_id ) {
			foreach ( $overwrite_data[ $main_snippet ] as $prop ) {
				if ( false !== stripos( $prop['snippet_id'], $snippet_id ) && ! $ends_with_fct( $prop['snippet_id'], $snippet_id ) ) {
					$d[ $prop['snippet_id'] ]['properties'][] = $prop;
					$d[ $prop['snippet_id'] ]['prop_name']    = $prop_name;
				}
			}
		}

		return $d;
	}


	/**
	 * Returns the HTML code for the properties to use in the table.
	 *
	 * Uses Rich_Snippet:get_properties() if $prop_ids has no elements.
	 *
	 * @param Rich_Snippet $snippet
	 * @param \WP_Post     $post
	 * @param int          $overwrite_post_id The post_id where the overwrite data is stored.
	 * @param string       $main_snippet_id
	 * @param string       $parent_snippet_id
	 *
	 * @return string[]
	 * @since 2.2.0 Added $post param.
	 * @since 2.14.0 Added $main_snippet_id param; removed $property_ids
	 *
	 * @since 2.0.0
	 */
	public function get_property_table_elements( $snippet, $post, $overwrite_post_id, $main_snippet_id, $parent_snippet_id ) {

		$overwrite_data = $this->get_properties_to_overwrite( $overwrite_post_id, $main_snippet_id, $snippet->id, $parent_snippet_id );

		if ( is_array( $overwrite_data ) && count( $overwrite_data ) > 0 ) {
			$snippet->overwrite_properties( $overwrite_data );
		}

		$list_data = $this->get_properties_to_list( $overwrite_post_id, $main_snippet_id, $snippet );
		if ( is_array( $list_data ) && count( $list_data ) > 0 ) {

			foreach ( $list_data as $snippet_id => $ld ) {
				$snippet->create_duplicate( $ld['prop_name'], $ld['properties'], $snippet_id );
			}
		}

		$html_elements = array();

		# load the properties from the snippet
		$props = $snippet->get_overridable_properties();

		# natural sort by id
		usort( $props, function ( $a, $b ) {
			return strcmp( $a->id, $b->id );
		} );

		foreach ( $props as $prop ) {
			if ( ! $prop instanceof Schema_Property ) {
				continue;
			}

			ob_start();
			View::admin_snippets_overwrite_row( $prop, $snippet, $post, $overwrite_post_id, $main_snippet_id, $parent_snippet_id );
			$html_elements[] = ob_get_clean();
		}

		if ( count( $html_elements ) <= 0 ) {
			$html_elements[] = sprintf( '<tr><td colspan="2">%s</td></tr>', __( 'No overridable properties found.', 'rich-snippets-schema' ) );
			$html_elements[] = sprintf(
				'<tr><td colspan="2"><img src="%s" alt="%s" width="809" height="573" /></td></tr>',
				plugins_url( '/img/make-them-editable.jpg', WPB_RS_FILE ),
				_x( 'How to make a property overridable.', 'Image alt text', 'rich-snippets-schema' )
			);
		}

		return $html_elements;
	}


	/**
	 * Builds a property table.
	 *
	 * Uses Rich_Snippet:get_properties() if $prop_ids has no elements.
	 *
	 * @param Rich_Snippet $snippet
	 * @param \WP_Post     $post
	 * @param int          $overwrite_post_id The post_id where the overwrite data is stored.
	 * @param string       $main_snippet_id
	 * @param string       $parent_snippet_id
	 *
	 * @return string
	 * @since 2.2.0 Added $post parameter
	 * @since 2.14.0 Added $main_snippet_id parameter; removed $prop_ids; Added $prop
	 *
	 * @since 2.0.0
	 */
	public function get_property_table( $snippet, $post, $overwrite_post_id, $main_snippet_id, $parent_snippet_id = '' ) {
		$props_rendered = $this->get_property_table_elements( $snippet, $post, $overwrite_post_id, $main_snippet_id, $parent_snippet_id );

		ob_start();

		View::admin_snippets_overwrite_table( $props_rendered, $snippet, $post );

		return ob_get_clean();
	}
}
